﻿using FukuracamDX;
using System.Collections.Generic;

namespace FatopTheWorld.Scenes{
	partial class Battle: FCAMScene{
		/// <summary>
		/// ScoreBoard は、そこにブロックがある時に加算（減算）
		/// 例）ぷよぷよの蛙戦法＆当然だが真ん中は避ける
		///		-14	-14	-14	-99	9	99	999
		///		-13	-13	-13	-99	9	99	999
		///		...	...	...	...	...	...	...
		///		-1	-1	-1	-99	9	99	999
		///		0	0	0	-99	9	99	999
		/// </summary>
		class CPUBrain{
			/// <summary>
			/// CPU計算時に保持する結果リストの最大数（ベース）。<br/>
			/// これは、ジュエルセットが3つあることで計算数が最大で３乗されるため、計算処理が遅くなるのを防ぐ制限値。<br/>
			/// なお省かれる時は、サイズの大きい順から省く。<br/>
			/// また再帰的に計算するたびに、この値は他の定数によって減算される。<br/>
			/// 最大数 = (MAX_CALCRESULT_LIST_BASE - MAX_CALCRESULT_LIST_DECREASE_STEPS) * MAX_CALCRESULT_LIST_DECREASE_RATE<br/>
			/// ※但しマクロの指定によっては再帰時の最大結果リスト数が固定される場合がある。
			/// </summary>
			private const int MAX_CALCRESULT_LIST_BASE = 4;
			/// <summary>CPU計算時の再帰計算で減算される結果リスト数（MAX_CALCRESULT_LIST_BASE参照）。</summary>
			private const int MAX_CALCRESULT_LIST_DECREASE_STEPS = 1;
			/// <summary>CPU計算時の再帰計算で乗算（1.0f未満）される結果リスト数（MAX_CALCRESULT_LIST_BASE参照）。</summary>
			private const double MAX_CALCRESULT_LIST_DECREASE_RATE = 0.5f;
			private FCAMRandom Random;
			/// <summary> 連鎖時に倍率に加算するステップ数。倍率は、消去したジュエル数の計算に対して乗算される。 </summary>
			public sbyte MultiplierSteps;
			/// <summary> ジュエルを消した時の得点 </summary>
			public int DeletePoints;
			/// <summary> 水平・垂直にジュエルを揃えた時の得点 </summary>
			public int HVPairPoints;
			/// <summary> 斜めにジュエルを揃えた時の得点 </summary>
			public int DiagonallyPairPoints;
			public CPUScoreBoard ScoreBoard;
			/// <summary>計算の深度。1が最低とし、2は1つ深読み、3が2つ深読み（深読みは、次の連鎖及び次のジュエルを総括）</summary>
			public sbyte Depth;
			/// <summary>精度。1 以上の値で、高いほど精度が落ちる(各々のCPUスコアを除算する)</summary>
			public short Precision;
			/// <summary>誤差。最終計算後、結果中での最大値からこの誤差分より小さいものは、候補から排除される。</summary>
			public short Rough;
			///	<summary>
			///	CPUが回転を行なう最大数。負の値を指定することで、CharacterInfo での難易度ごとのCPU生成時に
			///	回転数が変わらないようにさせられる（難易度が上がると計算式でそれぞれの情報が加減算されるため
			///	その際の対処法として負の値を指定する。）
			///	</summary>
			public sbyte MaxRotations;
			/// <summary> ダメージジュエルそのものの得点(基本負の値) </summary>
			public int DamageJewelPoints;
			public int HardDamageJewelPoints;
			public CPUDropType DropType;
			//	計算には使用しないがCPU処理として必要なもの
			/// <summary>CPUが動作を起こす感覚</summary>
			public int OperateInterval;
			/// <summary>CPUが動作を起こすまでの残りウェイト</summary>
			private int OperateWaitRemain;
			/// <summary>エナジーレベルのスコア</summary>
			public short EnergyLevelScore; 
			/// <summary>CPU計算時に敵のスコアを用いる時(例えば魔法で敵の動きを止める等)、そのスコアを割る数。</summary>
			public short OpponentCPUScoreDivide; 
			/// <summary>キャラクター固有(旧:攻撃)魔法での計算時に使用される、魔法ごとに異なる特別なポイント</summary>
			public short ForAttackMagicPoint = 100;
			/// <summary>補助魔法での計算時に使用される特別なポイント</summary>
			public int ForSupportMagicPoint = 1000;
			/// <summary>自分がダメージを受けている時に個々に加算されるポイント</summary>
			public int RecvDamagePoints;
			/// <summary>敵にダメージを与えている時に個々に加算されるポイント</summary>
			public int SendDamagePoints;
			/// <summary>敵にダメージを与えていない時に加算されるポイント。<br/>※相殺によってダメージが消された時にも加算されるので、RecvDamagePoints との兼ね合いが重要！</summary>
			public int NoSendDamagePoints;
			/// <summary>
			///	
			/// </summary>
			public CPUBrain(int operateInterval, sbyte maxRotations, sbyte depth, sbyte multiplierSteps, int recvDamagePoints, int sendDamagePoints, int noSendDamagePoints, CPUDropType dropType, short precision, short rough,
				int damageJewelPoints, int hardDamageJewelPoints, int deletePoints, int hvPairPoints, int diagonallyPairPoints, short energyLevelScore, short opponentCPUScoreDivide,
				CPUScoreBoard scoreBoard)
			{
				Random = new FCAMRandom();
				OperateInterval = operateInterval;
				MaxRotations = maxRotations;
				Depth = depth;
				MultiplierSteps = multiplierSteps;
				RecvDamagePoints = recvDamagePoints;
				SendDamagePoints = sendDamagePoints;
				NoSendDamagePoints = noSendDamagePoints;
				DropType = dropType;
				Precision = precision;
				Rough = rough;
				DamageJewelPoints = damageJewelPoints;
				HardDamageJewelPoints = hardDamageJewelPoints;
				DeletePoints = deletePoints;
				HVPairPoints = hvPairPoints;
				DiagonallyPairPoints = diagonallyPairPoints;
				EnergyLevelScore = energyLevelScore;
				OpponentCPUScoreDivide = opponentCPUScoreDivide;
				ScoreBoard = scoreBoard;
				ResetOperateWait();
			}
			public CPUBrain(int operateInterval, sbyte maxRotations, sbyte depth, sbyte multiplierSteps, int recvDamagePoints, int sendDamagePoints, int noSendDamagePoints, CPUDropType dropType, short precision, short rough,
				int damageJewelPoints, int hardDamageJewelPoints, int deletePoints, int hvPairPoints, int diagonallyPairPoints, short energyLevelScore, short opponentCPUScoreDivide,
				int[, ] cpuScore): this(
				operateInterval, maxRotations, depth, multiplierSteps, recvDamagePoints, sendDamagePoints, noSendDamagePoints, dropType, precision, rough,
				damageJewelPoints, hardDamageJewelPoints, deletePoints, hvPairPoints, diagonallyPairPoints, energyLevelScore, opponentCPUScoreDivide,
				new CPUScoreBoard(FIELD_TABLESIZE, cpuScore)
			){}
			public void ResetOperateWait(){
				OperateWaitRemain = OperateInterval;
			}
			/// <summary>
			/// 次の動作を行なう。もし操作待ちの動作なら false, 操作可能なら true を返す。
			/// </summary>
			public bool NextOperate(){
				return(--OperateWaitRemain <= 0);
			}
			/*
			public CPUBrain(){
				Random = new FCAMRandom();
				ScoreBoard = new CPUScoreBoard(FIELD_TABLESIZE);
				ScoreBoard.SetCPUScoreFromRotated(new int[, ]{
					{-0100, -0100, -0100, -9999, -0100,  0000,  0099},
					{-0100, -0100, -0100, -9999, -0100,  0000,  0099},
					{-0100, -0100, -0100, -9999, -0100,  0000,  0099},
					{-0100, -0100, -0100, -5000, -0100,  0000,  0099},
					{-0100, -0100, -0100, -5000, -0100,  0000,  0099},
					{-0100, -0100, -0100, -5000, -0100,  0000,  0099},
					{-0100, -0100, -0100, -1000, -0100,  0000,  0099},
					{-0100, -0100, -0100, -1000, -0100,  0000,  0099},
					{-0100, -0100, -0100, -1000, -0100,  0000,  0099},
					{-0100, -0100, -0100, -0500, -0100,  0000,  0099},
					{-0100, -0100, -0100, -0500, -0100,  0000,  0099},
					{-0100, -0100, -0100, -0500, -0100,  0000,  0099},
					{-0100, -0100, -0100, -0100, -0100,  0000,  0099},
					{-0100, -0100, -0100, -0100, -0100,  0000,  0099},
					{-0100, -0100, -0100, -0100, -0100,  0000,  0099},
				});
				MultiplierSteps = 5;
				DeletePoints = 10;
				HVPairPoints = 40;
				DiagonallyPairPoints = 80;
				Depth = 3;
				Precision = 10;
				Rough = 10;
				MaxRotations = 2;
				DamageJewelPoints = -30;
				UseInstantDrop = true;
			}
			*/
			/// <summary>
			/// CPUスコアを計算し、目標地点を計算。
			/// magic が true なら、魔法を計算に加える。
			/// 普通 magic は true だが、魔法を計算する際の ProcAndGetResult() で再び GetAim() を呼び出して計算することがあるので、
			/// そうすると無限ループに嵌ってしまうため、この引数がある。
			/// (即ち magic を false にするのは、基本魔法クラスでの上記 ProcAndGetResult() 関数のみ。)
			/// (但し魔法使用中は魔法が使えないので、プレイヤーがその状態の時のみは magic を false にする。)
			/// </summary>
			public CPUBrainCalcResult GetAim(Battle battle, Player player, bool magic, params Player[] targets){
				try{
#if TRACE
					FCAMDebug.Inform("--> GetAim()");
					FCAMDebug.StartNestedMeasuringTime();
#endif
					return(CalcCPUScore(battle,
						player.Field,
						magic,
#if ATTACK_MAGIC
						CharacterInfo.Infos[player.CharacterID, Game.MemoryData.Difficult].AttackMagic,
#endif
						CharacterInfo.Infos[player.CharacterID, Game.MemoryData.Difficult].SupportMagic,
						ScoreBoard,
						player.JewelSetList.Clone(),
						MaxRotations,
						Depth,
						player.Damage,
						player.DeleteLength,
						Precision,
						Rough,
						HVPairPoints,
						DiagonallyPairPoints,
						DeletePoints,
						MultiplierSteps,
						RecvDamagePoints,
						SendDamagePoints,
						NoSendDamagePoints,
						player.Energy,
						DamageJewelPoints,
						HardDamageJewelPoints,
						EnergyLevelScore,
						OpponentCPUScoreDivide,
						ForAttackMagicPoint,
						MAX_CALCRESULT_LIST_BASE,
						player,
						targets
					));
				}finally{
#if TRACE
					FCAMDebug.Inform("<-- GetAim():{0}", FCAMDebug.EndNestedMeasuringTimeString());
#endif
				}
			}
			/// <summary>
			///	指定 scoreBoard、PlayerField、depth と jewelSetList を元に、仮想上のフィールドで消去や落下を行なってCPUスコアを計算し、
			///	その複数の結果から一つを算出して CPUBrainCalcResult を返す。
			///	また depth は、連鎖とジュエルセットの先読み両方を兼ねている。
			///	例えば depth＝2 の時に現在のジュエルセットを配置した場合、
			///	連鎖が発生すれば2連鎖まで計算を行ない、
			///	もし連鎖なし（通常消去＝1連鎖 or 消去なし）の場合は、次のジュエルセットの配置（1連鎖まで）を計算する。
			///	<param name="fallSpeed">落下速度。速度によっては一番端まで移動できない場合があるので、それ用。</param>
			///	<param name="numberRotations">仮想フィールドで何回 JewelSet の回転パターンを取り入れるか。＝0：ぷよぷよのキャラ「ももも」風</param>
			/// <param name="precision">精度。1 以上の値で、高いほど精度が落ちる(各々のCPUスコアを除算する)</param>
			/// <param name="rough">大雑把さ。複数のCPUスコア計算結果から一つを選び出す際の、認められる範囲の CPUSScore 誤差([最大CPUスコア - rough]～[最大CPUスコア]</param>
			/// </summary>
			private static CPUBrainCalcResult CalcCPUScore(Battle battle,
				PlayerField field,
				bool magic,
#if ATTACK_MAGIC
				MagicProcessor attackMagic,
#endif
				MagicProcessor supportMagic,
				CPUScoreBoard scoreBoard,
				FCAMList<JewelSet> jewelSetList,
				int numberRotations,
				int depth,
				int damage,
				int deleteLength,
				int precision,
				int rough,
				int hvPairPoints,
				int diagonallyPairPoints,
				int deletePoints,
				int multiplierSteps,
				int recvDamagePoints,
				int sendDamagePoints,
				int noSendDamagePoints,
				int nowEnergy,
				int damageJewelPoints,
				int hardDamageJewelPoints,
				int EnergyLevelScore,
				int OpponentCPUScoreDivide,
				int ForAttackMagicPoint,
				int maxCalcResultList,
				Player player,
				params Player[] targets
			){
				if(depth <= 0 || jewelSetList.Count == 0 || maxCalcResultList <= 0){
					return(new CPUBrainCalcResult());
				}
				long maxCPUScore;
				JewelSet jewelSet;
				List<CPUBrainCalcResult> calcResultList;
#if TRACE
				FCAMDebug.Inform("--> CalcCPUScore():depth={0},jewelSetList.Length={1}", depth, jewelSetList.Count);
#endif
				//	ジュエルセットをそれぞれの位置に配置した時の CPUScore を算出する。
				calcResultList = new List<CPUBrainCalcResult>();
				jewelSet = jewelSetList[0].Clone();
				maxCPUScore = long.MinValue;
				//	レベルが変わっても絶対回転させないよう numberRotation を負にしている場合があるため、その対処ようとして最低でも 0 には設定させる
				//	また最大回転数として 99 を指定した場合、99回回転させてしまうことがあるため、
				//	回転数を最大とするよう設定する。
				//	なお先に Min からしないと、ありえないがもし Rows が 0 の時 -1 が採用されておかしくなる。
				numberRotations = FCAMMath.Max(FCAMMath.Min(numberRotations, jewelSet.Size.Rows - 1), 0);
				//	計算開始
				foreach(MagicKind kind in new[]{
					MagicKind.Nothing,
#if ATTACK_MAGIC
					MagicKind.Attack,
#endif
					MagicKind.Support
				}){
					if(kind != MagicKind.Nothing){
						if(!magic || Player.GetEnergyLevel(nowEnergy) <= 0){
							continue;
						}
					}
					for(sbyte rotations = 0; rotations <= numberRotations; rotations++){
						int column = JEWEL_SET_START_POS_COLUMN;
						int startColumn = player.MoveRange.StartColumn;
						int endColumn = player.MoveRange.EndColumn;
						{
							int fallSpeed = battle.GetFallSpeed(player);
							int putTime = battle.GetPutTime(player);
							//	ここでの row は、ジュエルセットの一番下を基準とする。
							int row = FIELD_BLIND_HEIGHT;
							int shiftY = 0;
							int nowPutWait = 0;
							int operateWaitRemain = player.CPU.OperateInterval;
							//	本来なら一番外側のネスト？で計算すれば、一律同じ値（カラム位置と可動範囲）で初期化出来るが、
							//	もし仮に rotation が並び替えではなく本当に回転を行なう場合は、
							//	column などの上方を再計算する必要があるため、計算数は倍増するがこの位置で column などを初期化する。
							//	まずは左へどこまで行けるのかを算出（スタート位置 startColumn の決定）。
							while(nowPutWait <= putTime
							   && column > startColumn
							   && field.Data[column - 1, row].Kind == 0){
								operateWaitRemain--;
								if(operateWaitRemain <= 0){
									column--;
									operateWaitRemain = player.CPU.OperateInterval;
								}
								shiftY += fallSpeed;
								row += shiftY / JEWEL_SIZE.Height;
								shiftY %= JEWEL_SIZE.Height;
								if(row >= field.Size.Rows - 1){
									row = field.Size.Rows - 1;
									//	一回の移動ですぐに横へ移動が完了するわけではないため、
									//	追加される待ち時間は移動量を元に算出する。
									nowPutWait += JEWEL_SIZE.Width / JEWEL_HORZ_MOVE_SPEED;
								}
							}
							startColumn = column;
							//	次に右へどこまで行けるのかを算出（エンド位置 endColumn の決定）。
							column = JEWEL_SET_START_POS_COLUMN;
							row = FIELD_BLIND_HEIGHT;
							shiftY = 0;
							nowPutWait = 0;
							operateWaitRemain = player.CPU.OperateInterval;
							while(nowPutWait <= putTime
							   && column < endColumn - 1
							   && field.Data[column + 1, row].Kind == 0){
								operateWaitRemain--;
								if(operateWaitRemain <= 0){
									column++;
									operateWaitRemain = player.CPU.OperateInterval;
								}
								shiftY += fallSpeed;
								row += shiftY / JEWEL_SIZE.Height;
								shiftY %= JEWEL_SIZE.Height;
								if(row >= field.Size.Rows - 1){
									row = field.Size.Rows - 1;
									//	一回の移動ですぐに横へ移動が完了するわけではないため、
									//	追加される待ち時間は移動量を元に算出する。
									nowPutWait += JEWEL_SIZE.Width / JEWEL_HORZ_MOVE_SPEED;
								}
							}
							endColumn = column;
						}
						column = startColumn;
						do{
							int multiplier;
							DeleteField deleteField;
							CPUBrainCalcResult calcResult = new CPUBrainCalcResult();
							calcResult.Depth = depth;
							calcResult.Point.Column = column;
							calcResult.Point.Row = player.MoveRange.Row;
							calcResult.Rotations = rotations;
							calcResult.UseMagic = kind;
							calcResult.CloneField = field.Clone();
							calcResult.CPUScore = 0;
							calcResult.Jewels = 0;
							calcResult.Chain = 0;
							multiplier = 1;
							calcResult.Energy = nowEnergy;
							//	魔法を使用するならしてみる（仮想空間で）
							if(Player.GetEnergyLevel(calcResult.Energy) > 0){
								switch(kind){
									case MagicKind.Nothing:{
										//	魔法がない時は何もしない。
										break;
									}
#if ATTACK_MAGIC
									case MagicKind.Attack:{
										calcResult.CPUScore += attackMagic.ProcAndGetResult(ref calcResult.CloneField, ref jewelSetList, player, column, targets);
										calcResult.Energy = Player.GetEnergyShiftedByLevel(calcResult.Energy, attackMagic.GetNeedLevel());
										break;
									}
#endif
									case MagicKind.Support:{
										calcResult.CPUScore += supportMagic.ProcAndGetResult(ref calcResult.CloneField, ref jewelSetList, player, column, targets);
										calcResult.Energy = Player.GetEnergyShiftedByLevel(calcResult.Energy, supportMagic.GetNeedLevel());
										break;
									}
									default:{
										Error.ID_00001.Throw();
										break;
									}
								}
							}
							//	宙に浮いた状態で現在位置にジュエルセットを配置。
							calcResult.CloneField.PutJewelSet(jewelSet, column, FIELD_BLIND_HEIGHT);
							//	宙に浮いた状態でジュエルセットを設置したので、それを一番下まで落とす。
							FallFloatingJewels(calcResult.CloneField, true);
							//	＜深読み＞連鎖を行なう。ここでも depth は消費する。
							//	連鎖でまずは depth を消費。その消費後 depth を再帰呼び出しの CalcCPUScore() に渡す。
							do{
								calcResult.Depth--;
								deleteField = GetDeleteField(calcResult.CloneField, deleteLength, calcResult.Chain + 1);
								if(!deleteField.ExistDeleteJewel())
									break;
								int recover = FCAMMath.Min(damage, deleteField.Damage);
								damage -= recover;
								deleteField.Damage -= recover;
								calcResult.Jewels += DeleteJewelsIn(calcResult.CloneField, deleteField);
								FallFloatingJewels(calcResult.CloneField, true);
								multiplier += multiplierSteps;
								calcResult.Chain++;
								calcResult.Energy += deleteField.Energy;
							}while(calcResult.Depth > 0);
							//	消去処理を行なったあとの処理
							//	普通に配置した場合にゲームオーバーになるのだったら
							if(calcResult.CloneField.Data[JEWEL_DEAD_POINT.Column, JEWEL_DEAD_POINT.Row].Kind > 0){
								calcResult.PinchLevel = CPUPinchLevel.Absolute;
							}
							//	まずはお邪魔を落とす(この時 PC補正で先読みさせない――というかそういう処理は出来ない――ので、予想で考える)
							{
								//	小数点以下繰り上げ（%での計算の所）で行数を計算
								//	これにより、仮に1行分に満たないお邪魔でも、何処に落ちても良いような予想を取る
								for(int rows = (damage / field.Size.Columns) + (damage % field.Size.Columns > 0 ? 1 : 0); rows > 0; rows--){
									for(int i = field.Size.Columns - 1; i >= 0; i--){
										if(calcResult.CloneField.Data[i, 0].Kind == 0){
											calcResult.CloneField.Data[i, 0].Put(OBSTACLE_JEWEL);
											damage--;
										}
									}
									//	一番下まで落下
									FallFloatingJewels(calcResult.CloneField, true);
								}
								//	ダメージの計算を終えたら 0 にする。
								//damage = 0;
								//	＜上記＞でも 0 にしたら、深読みの際にダメージがリセットされるから、正しく計算されない？
								//	なのでお邪魔ジュエル配置時にダメージを引いて、必要なら残すようにする。

								//	お邪魔の落下予想でゲームオーバーになるのだったら
								if(calcResult.PinchLevel != CPUPinchLevel.Absolute && calcResult.CloneField.Data[JEWEL_DEAD_POINT.Column, JEWEL_DEAD_POINT.Row].Kind > 0){
									if(damage == 0){
										//	ダメージが残っていなければ予想でゲームオーバー
										calcResult.PinchLevel = CPUPinchLevel.Probability;
									}else{
										//	ダメージが残っていたら、確実にゲームオーバー
										calcResult.PinchLevel = CPUPinchLevel.Absolute;
									}
								}
							}
							//	CPUScore とフィールドの CPUScore などを加算。
							calcResult.CPUScore += Player.GetTrueEnergy(calcResult.Energy) + Player.GetEnergyLevel(calcResult.Energy) * EnergyLevelScore;
							calcResult.CPUScore += damage * recvDamagePoints;
							calcResult.CPUScore += deleteField.Damage * sendDamagePoints;
							if(deleteField.Damage <= 0){
								calcResult.CPUScore += noSendDamagePoints;
							}
							calcResult.CPUScore += deleteField.HVPairJewels * hvPairPoints + deleteField.DiagonallyPairJewels * diagonallyPairPoints;
							calcResult.CPUScore += calcResult.Jewels * deletePoints * multiplier;
							calcResult.CPUScore += damageJewelPoints * deleteField.DamageJewels;
							calcResult.CPUScore += hardDamageJewelPoints * deleteField.HardDamageJewels;
							calcResult.CPUScore += GetFieldCPUScore(calcResult.CloneField, scoreBoard);
							calcResult.CPUScore /= precision;
							//	＜深読み＞次のジュエル（前バージョン：絶対再帰計算）
#if CPU_CALC_RECURSIVE_ABSOLUTE
							//if(__depth > 1 && jewelSetList.Count > 1){
								var recursiveResult = CalcCPUScore(magic,
									calcResult.CloneField,
	#if ATTACK_MAGIC
									attackMagic,
	#endif
									supportMagic,
									scoreBoard,
									jewelSetList.Clone(1),
									numberRotations,
									calcResult.Depth - 1,
									damage,
									deleteLength,
									precision,
									rough,
									hvPairPoints,
									diagonallyPairPoints,
									deletePoints,
									multiplierSteps,
									calcResult.Energy,
									damageJewelPoints,
									hardDamageJewelPoints,
									EnergyLevelScore,
									OpponentCPUScoreDivide,
									ForAttackMagicPoint,
	#if CPU_CALC_RECURSIVE_SIMPLE
									1,
	#else
									(int)(MAX_CALCRESULT_LIST_DECREASE_RATE * (maxCalcResultList - MAX_CALCRESULT_LIST_DECREASE_STEPS)),
	#endif
									player,
									targets
								);
								//
								calcResult.CPUScore += recursiveResult.CPUScore;
								calcResult.Jewels += recursiveResult.Jewels;
								calcResult.Chain += recursiveResult.Chain;
							//}
#endif
							//
							maxCPUScore = FCAMMath.Max(maxCPUScore, calcResult.CPUScore);
							calcResultList.Add(calcResult);
							column++;
							//	while の比較は field。何故なら cloneField は仮想フィールドのため、実際に配置して見る位置とは異なるから
						}while(column <= endColumn && field.Data[column, FIELD_BLIND_HEIGHT].Kind == 0);
						jewelSet.Rotate();
					}
				}
				//
#if DEBUG
				FCAMDebug.NOP();
#endif
				//	大雑把さの範囲外にある計算結果は削除する。
				//	後ろから処理を行なうのは、削除した際に後ろの方からインデックスがずれるので、おかしくなる為。
#if TRACE
				FCAMDebug.Inform("calcResultList.Count:1/4-Before = {0}", calcResultList.Count);
#endif
				for(int i = calcResultList.Count - 1; i >= 0; i--){
					if(calcResultList[i].CPUScore < maxCPUScore - rough){
						calcResultList.RemoveAt(i);
					}
				}
				//	計算結果を削除したあとも、結果数が最大値を超えて居た場合は更に削除する。
#if TRACE
				FCAMDebug.Inform("calcResultList.Count:2/4-Limit = {0}", calcResultList.Count);
#endif
				while(calcResultList.Count > maxCalcResultList){
					//	一番小さいスコアから順々に削除する。
					long min = long.MaxValue;
					int minIndex = -1;
					for(int i = calcResultList.Count - 1; i >= 0; i--){
						if(calcResultList[i].CPUScore < min){
							min = calcResultList[i].CPUScore;
							minIndex = i;
						}
					}
					//	minIndex が -1 だとプログラムミスであるため、ここではそれに対する分岐処理はしない。
					calcResultList.RemoveAt(minIndex);
				}
#if !CPU_CALC_RECURSIVE_ABSOLUTE
	#if TRACE
				FCAMDebug.Inform("RECURSIVE:calcResultList={0}", calcResultList.Count);
	#endif
				//	再帰計算でスコアも再計算されるため、最大スコア値を初期化する。
				maxCPUScore = long.MinValue;
				//	＜深読み＞次のジュエル（後バージョン）
				foreach(CPUBrainCalcResult elem in calcResultList){
					var recursiveResult = CalcCPUScore(battle,
						elem.CloneField,
						magic,
#if ATTACK_MAGIC
						attackMagic,
#endif
						supportMagic,
						scoreBoard,
						jewelSetList.Clone(1),
						numberRotations,
						elem.Depth - 1,
						damage,
						deleteLength,
						precision,
						rough,
						hvPairPoints,
						diagonallyPairPoints,
						deletePoints,
						multiplierSteps,
						recvDamagePoints,
						sendDamagePoints,
						noSendDamagePoints,
						elem.Energy,
						damageJewelPoints,
						hardDamageJewelPoints,
						EnergyLevelScore,
						OpponentCPUScoreDivide,
						ForAttackMagicPoint,
		#if CPU_CALC_RECURSIVE_SIMPLE
						1,
		#else
						(int)(MAX_CALCRESULT_LIST_DECREASE_RATE * (maxCalcResultList - MAX_CALCRESULT_LIST_DECREASE_STEPS)),
		#endif
						player,
						targets
					);
					//
					elem.CPUScore += recursiveResult.CPUScore;
					elem.Jewels += recursiveResult.Jewels;
					elem.Chain += recursiveResult.Chain;
					maxCPUScore = FCAMMath.Max(maxCPUScore, elem.CPUScore);
				}
				//	再度誤差範囲チェック＆削除を行なう。
				//	但し計算結果数は前の処理で限界点を定めており、ここでは追加されることはないので
				//	純粋に範囲チェックによる削除のみでＯＫ。
	#if TRACE
				FCAMDebug.Inform("calcResultList.Count:3/4-After = {0}", calcResultList.Count);
	#endif
				for(int i = calcResultList.Count - 1; i >= 0; i--){
					if(calcResultList[i].CPUScore < maxCPUScore - rough){
						calcResultList.RemoveAt(i);
					}
				}
#endif
				//	計算結果リストの中からランダムで目標地点を返却する。
				try{
#if TRACE
					FCAMDebug.Inform("calcResultList.Count:4/4-Final = {0}", calcResultList.Count);
#endif
					//if(calcResultList.Count > 0){
						return(calcResultList[FCAMRandom.Common.NextInt(calcResultList.Count)]);
					//}else{
						//	計算結果リスト数が 0 件はありえないことだが、とりあえずその場合は左右どちらかの端を設定。
						//var returnValue = new CPUBrainCalcResult();
						//returnValue.Chain = 0;
						//returnValue.CloneField = null;
						//returnValue.CPUScore = 0;
						//returnValue.Depth = 0;
						//returnValue.Energy = 0;
						//returnValue.Jewels = 0;
						//returnValue.PinchLevel = CPUPinchLevel.Nothing;
						//returnValue.Rotations = 0;
						//returnValue.UseMagic = MagicKind.Nothing;
						//if(FCAMRandom.Common.NextBool()){
						//    returnValue.Point.Set(0, 0);
						//}else{
						//    returnValue.Point.Set(field.Size.Columns - 1, 0);
						//}
						//return returnValue;
					//}
				}finally{
#if TRACE
				FCAMDebug.Inform("<-- CalcCPUScore()");
#endif
				}
			}
			/// <summary>
			///	指定 PlayerField のCPUスコアを取得する。
			/// </summary>
			private static long GetFieldCPUScore(PlayerField field, CPUScoreBoard scoreBoard){
				long cpuScore = 0;
				for(int row = 0; row < field.Size.Rows; row++){
					for(int column = 0; column < field.Size.Columns; column++){
						if(field.Data[column, row].Kind != 0)
							cpuScore += scoreBoard.Data[column, row];
					}
				}
				return(cpuScore);
			}
		}
		/// <summary>
		/// 
		/// </summary>
		class CPUScoreBoard: Board<int>{
			/// <summary>致死エリアを示すためのスコアポイント。-99999</summary>
			public const int DEAD_AREA_POINT = -99999;
			/// <summary>
			/// 通常コンストラクタ
			/// </summary>
			public CPUScoreBoard(FCAMTableSize size): base(size){}
			/// <summary>
			///	スコアボードを自動計算するコンストラクタ
			/// </summary>
			public CPUScoreBoard(FCAMTableSize size, int topLeft, int topCenter, int topRight, int middleCenter, int bottomLeft, int bottomCenter, int bottomRight): base(size){
				int halfColumns = size.Columns / 2;
				int halfRows = size.Rows / 2;
				int maxRow = size.Rows - 1;
				for(int column = 0; column < size.Columns; column++){
					//	一番上と一番下の行のスコアを計算する。
					for(int row = 0; row < size.Rows; row += maxRow){
						int left, right, number, length;
						if(column < halfColumns){
							number = column;
							length = halfColumns;
							if(row > 0){
								left = bottomLeft;
								right = bottomCenter;
							}else{
								left = topLeft;
								right = topCenter;
							}
						}else if(column > halfColumns){
							number = column - halfColumns;
							length = halfColumns;
							if(row > 0){
								left = bottomCenter;
								right = bottomRight;
							}else{
								left = topCenter;
								right = topRight;
							}
						}else{
							number = length = 1;
							left = right = row > 0 ? bottomCenter : topCenter;
						}
						Data[column, row] = (right - left) * number / length + left;
					}
					//	一番下と一番上のスコアを元に間のスコアを計算する。
					for(int row = 1; row < maxRow; row++){
						int top, bottom, number, length;
						if(column < halfColumns){
							number = row;
							length = maxRow;
							top = Data[column, 0];
							bottom = Data[column, maxRow];
						}else if(column > halfColumns){
							number = row;
							length = maxRow;
							top = Data[column, 0];
							bottom = Data[column, maxRow];
						}else{
							if(row < halfRows){
								number = row;
								length = halfRows;
								top = Data[column, 0];
								bottom = middleCenter;
							}else if(row > halfRows){
								number = row - halfRows;
								length = halfRows;
								top = middleCenter;
								bottom = Data[column, maxRow];
							}else{
								number = length = 1;
								top = bottom = middleCenter;
							}
						}
						Data[column, row] = (bottom - top) * number / length + top;
					}
				}
				//	致死エリアには置かないようなスコア配置
				for(int row = 0; row < JEWEL_SET_SIZE.Rows + FIELD_BLIND_HEIGHT; row++){
					Data[JEWEL_SET_START_POS_COLUMN, row] = DEAD_AREA_POINT;
				}
			}
			/// <summary>
			///	スコアボードを配列の1次元と2次元を逆にさせて設定するコンストラクタ
			/// </summary>
			public CPUScoreBoard(FCAMTableSize size, int[, ] data): base(size){
				for(int row = 0; row < Size.Rows; row++){
					for(int column = 0; column < Size.Columns; column++){
						Data[column, row] = data[row, column];
					}
				}
			}
			/// <summary>
			/// 
			/// </summary>
			public override string ToString(){
				System.Text.StringBuilder sb = new System.Text.StringBuilder();
				sb.Append(FCAMSystem.NewLine);
				for(int row = 0; row < Size.Rows; row++){
					for(int column = 0; column < Size.Columns; column++){
						sb.Append(FCAMStrings.Format("{0,7}, ", Data[column, row]));
					}
					sb.Append(FCAMSystem.NewLine);
				}
				return(sb.ToString());
			}
		}
		/// <summary>
		/// CPUBrain のCPUスコア計算結果
		/// </summary>
		class CPUBrainCalcResult{
#if !CPU_CALC_RECURSIVE_ABSOLUTE
			/// <summary>この状態での仮想フィールド</summary>
			public PlayerField CloneField = null;
			/// <summary>この状態での残り深度</summary>
			public int Depth = -1;
			/// <summary>この状態でのエナジー量</summary>
			public int Energy = -1;
#endif
			/// <summary> CPUスコアを叩き出した位置 </summary>
			public FCAMTablePoint Point = new FCAMTablePoint(-1);
			/// <summary> CPUスコアを叩き出したジュエルセットの回転数 </summary>
			public sbyte Rotations = -1;
			/// <summary>使用する魔法</summary>
			public MagicKind UseMagic = MagicKind.Nothing;
			/// <summary> CPUスコア </summary>
			public long CPUScore = 0;
			/// <summary>消去ジュエル数（合計）</summary>
			public int Jewels = 0;
			/// <summary>発生連鎖数（合計）</summary>
			public int Chain = 0;
			/// <summary>指定の位置に配置したらゲームオーバーになるかなどのピンチ度</summary>
			public CPUPinchLevel PinchLevel = CPUPinchLevel.Nothing;
			/// <summary>
			/// 
			/// </summary>
			/// <returns></returns>
			public override string ToString() {
				return FCAMStrings.Format("Score:{0,16} Point:{{{1}}} Rotate:{2} Magic:{3} Chain:{4} Jewels:{5}", CPUScore, Point, Rotations, UseMagic, Chain, Jewels);
			}
		}
		/// <summary>
		/// CPUの落下の種類
		/// </summary>
		public enum CPUDropType: sbyte{
			/// <summary>自ら落下はしない。</summary>
			None = 0,
			/// <summary>自ら通常落下はするが、ダメージが致死の場合は落下しない。</summary>
			Normal,
			/// <summary>自ら高速落下はするが、ダメージが致死の可能性有なら通常落下で、ダメージが致死の場合は落下しない。</summary>
			Instant,
			MAX = Instant
		}
		/// <summary>CPU計算時に割り出した位置でのピンチ度</summary>
		enum CPUPinchLevel{
			Nothing,
			/// <summary>お邪魔の落下の運次第でゲームオーバー</summary>
			Probability,
			/// <summary>確実にゲームオーバー</summary>
			Absolute,
		}
	}
}
